Imports EnvDTE
Imports EnvDTE80

Public Class dlgAddDelegationCode
    ' The add-in must set these before displaying the dialog.
    Public ApplicationObject As DTE2
    Public DelegatingClass As CodeClass2

    ' Stores a name/value pair.
    Private Class NameValuePair
        Public Name As String
        Public Value As Object
        Public Overrides Function ToString() As String
            Return Name
        End Function
        Public Sub New(ByVal new_name As String, ByVal new_value As Object)
            Name = new_name
            Value = new_value
        End Sub
    End Class

    ' Prepare the dialog.
    Private Sub dlgAddDelegationCode_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        ' Loop through all solutions.
        For Each proj As Project In ApplicationObject.Solution.Projects
            ' Loop through project items in the project.
            For Each project_item As ProjectItem In proj.ProjectItems
                ' See if this project item has a code model.
                ' (For example, My Project does not.)
                If project_item.FileCodeModel IsNot Nothing Then
                    ' Loop through the project item's code elements.
                    Dim fcm As FileCodeModel2 = DirectCast(project_item.FileCodeModel, FileCodeModel2)
                    For Each code_element As CodeElement2 In fcm.CodeElements
                        cboClass.Items.Add(New NameValuePair(code_element.Name, code_element))
                    Next code_element
                End If
            Next project_item
        Next proj
    End Sub

    ' Display properties and methods provided by this class.
    Private Sub cboClass_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cboClass.SelectedIndexChanged
        btnOk.Enabled = True

        ' Get the selected object.
        Dim name_value_pair As NameValuePair = DirectCast(cboClass.SelectedItem, NameValuePair)
        Dim class_element As CodeElement2 = DirectCast(name_value_pair.Value, CodeElement2)

        ' See what items this class provides.
        lstItems.Items.Clear()
        For Each code_element As CodeElement2 In class_element.Children()
            Dim txt As String = ""

            ' Only list Public and Friend items.
            Select Case code_element.Kind
                Case EnvDTE.vsCMElement.vsCMElementEvent
                    Dim code_event As CodeEvent = DirectCast(code_element, CodeEvent)
                    txt = DeclarationEvent(code_element)

                Case EnvDTE.vsCMElement.vsCMElementFunction
                    Dim code_function As CodeFunction2 = DirectCast(code_element, CodeFunction2)
                    If code_function.Type.TypeKind = vsCMTypeRef.vsCMTypeRefVoid Then
                        txt = DeclarationSub(code_function)
                    Else
                        txt = DeclarationFunction(code_function)
                    End If

                Case EnvDTE.vsCMElement.vsCMElementProperty
                    Dim code_property As CodeProperty2 = DirectCast(code_element, CodeProperty2)
                    txt = DeclarationProperty(code_property)

                Case EnvDTE.vsCMElement.vsCMElementVariable
                    Dim code_variable As CodeVariable2 = DirectCast(code_element, CodeVariable2)
                    txt = DeclarationVariable(code_variable)
            End Select

            ' Ignore any that are not Public or Friend.
            If txt.Length > 0 Then
                lstItems.Items.Add(New NameValuePair(txt, code_element))

                ' Don't check constructors.
                If code_element.Name <> "New" Then
                    lstItems.SetItemChecked(lstItems.Items.Count - 1, True)
                End If
            End If
        Next code_element
    End Sub

    ' Return a string representing the object's access.
    Private Function AccessAndPrototype(ByVal access_type As vsCMAccess, ByVal code_type As String, ByVal prototype As String) As String
        Dim txt As String = ""

        Select Case access_type
            Case vsCMAccess.vsCMAccessPublic
                txt = "Public " & code_type & prototype
            Case vsCMAccess.vsCMAccessProject
                txt = "Friend " & code_type & prototype
            Case Else
                ' We can't use Private, Protected, or Protected Friend.
                txt = ""
        End Select

        Return txt
    End Function

    ' Build the delegating object.
    Private Sub btnOk_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnOk.Click
        ' Get the selected class.
        Dim class_name_value_pair As NameValuePair = DirectCast(cboClass.SelectedItem, NameValuePair)
        Dim class_element As CodeElement2 = DirectCast(class_name_value_pair.Value, CodeElement2)

        ' Declare the variable.
        Dim object_name As String = "m_" & class_element.Name
        Dim txt As String = "#Region ""Delegated to " & object_name & """" & vbCrLf
        txt &= vbTab & "Private WithEvents " & object_name & " As " & class_element.Name & vbCrLf & vbCrLf

        ' Loop through the selected methods.
        For Each name_value_pair As NameValuePair In lstItems.CheckedItems
            ' Get the code element.
            Dim code_element As CodeElement2 = DirectCast(name_value_pair.Value, CodeElement2)

            ' Make code for the code element.
            Select Case code_element.Kind
                Case vsCMElement.vsCMElementEvent
                    txt &= ImplementationForEvent(object_name, code_element)

                Case vsCMElement.vsCMElementFunction
                    txt &= ImplementationForFunction(object_name, code_element)

                Case vsCMElement.vsCMElementProperty
                    txt &= ImplementationForProperty(object_name, code_element)

                Case vsCMElement.vsCMElementVariable
                    txt &= ImplementationForVariable(object_name, code_element)

            End Select
        Next name_value_pair

        ' End the region.
        txt &= "#End Region ' Delegated to " & object_name & vbCrLf

        ' Get the start of the delgating class's code.
        Dim text_point As TextPoint = DelegatingClass.GetStartPoint(vsCMPart.vsCMPartBody)
        Dim edit_point As EditPoint = text_point.CreateEditPoint()
        edit_point.Insert(txt)
    End Sub

#Region "Declarations for code elements"
    ' Generate declarations for various code elements.
    Private Function DeclarationEvent(ByVal code_element As CodeElement2) As String
        Dim code_event As CodeEvent = DirectCast(code_element, CodeEvent)
        Return AccessAndPrototype( _
            code_event.Access, _
            "Event ", _
            code_event.Prototype( _
                vsCMPrototype.vsCMPrototypeParamNames Or _
                vsCMPrototype.vsCMPrototypeParamTypes Or _
                vsCMPrototype.vsCMPrototypeType))
    End Function

    Private Function DeclarationSub(ByVal code_function As CodeFunction2) As String
        Return AccessAndPrototype( _
            code_function.Access, _
            "Sub ", _
            code_function.Prototype( _
                vsCMPrototype.vsCMPrototypeParamNames Or _
                vsCMPrototype.vsCMPrototypeParamTypes Or _
                vsCMPrototype.vsCMPrototypeType))
    End Function

    Private Function DeclarationFunction(ByVal code_function As CodeFunction2) As String
        Return AccessAndPrototype( _
            code_function.Access, _
            "Function ", _
            code_function.Prototype( _
                vsCMPrototype.vsCMPrototypeParamNames Or _
                vsCMPrototype.vsCMPrototypeParamTypes Or _
                vsCMPrototype.vsCMPrototypeType))
    End Function

    Private Function DeclarationProperty(ByVal code_property As CodeProperty2) As String
        ' Look for the property get and set procedures.
        Dim is_readonly As Boolean = True
        Dim is_writeonly As Boolean = True
        Dim data_type As String = ""
        For i As Integer = 1 To code_property.Children.Count
            Dim code_function As CodeFunction2 = DirectCast(code_property.Children.Item(i), CodeFunction2)
            If code_function.FunctionKind = vsCMFunction.vsCMFunctionPropertySet Then
                ' This is the property set procedure.
                is_readonly = False
                Dim code_parameter As CodeParameter2 = DirectCast(code_function.Parameters.Item(code_function.Parameters.Count), CodeParameter2)

                ' Ignore System.Void.
                If code_parameter.Type.TypeKind <> vsCMTypeRef.vsCMTypeRefVoid Then data_type = code_parameter.Type.AsString
            ElseIf code_function.FunctionKind = vsCMFunction.vsCMFunctionPropertyGet Then
                ' This is the property get procedure.
                is_writeonly = False
                ' Check for Length = 0 because sometimes 

                ' Ignore System.Void.
                If code_function.Type.TypeKind <> vsCMTypeRef.vsCMTypeRefVoid Then data_type = code_function.Type.AsString
            Else
                Stop
            End If
        Next i

        ' See if it's ReadOnly or WriteOnly.
        Dim read_write As String
        If is_readonly Then
            read_write = "ReadOnly "
        ElseIf is_writeonly Then
            read_write = "WriteOnly "
        Else
            read_write = ""
        End If

        ' Compose the declaration.
        Dim txt As String = AccessAndPrototype( _
            code_property.Access, _
            read_write & "Property ", _
            code_property.Prototype( _
                vsCMPrototype.vsCMPrototypeParamNames Or _
                vsCMPrototype.vsCMPrototypeParamTypes Or _
                vsCMPrototype.vsCMPrototypeType))

        ' Add the data type.
        txt &= " As " & data_type
        Return txt
    End Function

    Private Function DeclarationVariable(ByVal code_variable As CodeVariable2) As String
        Dim access_type As vsCMAccess = code_variable.Access
        If (access_type = vsCMAccess.vsCMAccessPublic) OrElse _
           (access_type = vsCMAccess.vsCMAccessProject) _
        Then
            ' It's Public or Friend. Use it.
            Return code_variable.Prototype(vsCMPrototype.vsCMPrototypeType)
        Else
            ' It's not Public or Friend. Skip it.
            Return ""
        End If
    End Function
#End Region ' Declarations for code elements

    ' Return the name for this access type (Public, Friend, etc.).
    Private Function AccessName(ByVal access As vsCMAccess) As String
        Select Case access
            Case vsCMAccess.vsCMAccessPublic
                Return "Public"
            Case vsCMAccess.vsCMAccessProject
                Return "Friend"
            Case Else
                Return ""
        End Select
    End Function

#Region "Implementations for code elements"
    ' Return code to implement code elements.
    Private Function ImplementationForEvent(ByVal object_name As String, ByVal code_element As CodeElement2) As String
        Dim txt As String = ""

        ' Declare the event.
        txt &= vbTab & DeclarationEvent(code_element) & vbCrLf

        ' Start the event handler.
        txt &= vbTab & "Private Sub " & object_name & "_"
        Dim code_event As CodeEvent = DirectCast(code_element, CodeEvent)
        txt &= code_event.Prototype( _
            vsCMPrototype.vsCMPrototypeParamNames Or _
            vsCMPrototype.vsCMPrototypeParamTypes)
        txt &= " Handles " & object_name & "." & code_event.Name & vbCrLf

        ' Delegate.
        Dim params As String = code_event.Prototype(vsCMPrototype.vsCMPrototypeParamNames)
        params = params.Replace("ByVal ", "")
        params = params.Replace("ByRef ", "")
        txt &= vbTab & vbTab & "RaiseEvent " & params & vbCrLf

        txt &= vbTab & "End Sub" & vbCrLf & vbCrLf
        Return txt
    End Function

    Private Function ImplementationForFunction(ByVal object_name As String, ByVal code_element As CodeElement2) As String
        Dim txt As String = ""

        Dim code_function As CodeFunction2 = DirectCast(code_element, CodeFunction2)
        If code_function.Type.TypeKind = vsCMTypeRef.vsCMTypeRefVoid Then
            ' It's a subroutine.
            ' Start the subroutine.
            txt &= vbTab & DeclarationSub(code_function) & vbCrLf

            ' Delegate.
            Dim params As String = code_function.Prototype(vsCMPrototype.vsCMPrototypeParamNames)
            params = params.Replace("ByVal ", "")
            params = params.Replace("ByRef ", "")
            params = params.Replace(" ,", ",")
            params = params.Replace(" )", ")")
            txt &= vbTab & vbTab & params & vbCrLf

            txt &= vbTab & "End Sub" & vbCrLf & vbCrLf
        Else
            ' It's a function.
            ' Start the function.
            txt &= vbTab & DeclarationFunction(code_function) & vbCrLf

            ' Delegate.
            Dim params As String = code_function.Prototype(vsCMPrototype.vsCMPrototypeParamNames)
            params = params.Replace("ByVal ", "")
            params = params.Replace("ByRef ", "")
            params = params.Replace(" ,", ",")
            params = params.Replace(" )", ")")
            txt &= vbTab & vbTab & "Return " & object_name & "." & params & vbCrLf

            txt &= vbTab & "End Function" & vbCrLf & vbCrLf
        End If

        Return txt
    End Function

    Private Function ImplementationForProperty(ByVal object_name As String, ByVal code_element As CodeElement2) As String
        Dim txt As String = ""

        ' Start the property procedure.
        Dim code_property As CodeProperty2 = DirectCast(code_element, CodeProperty2)
        txt &= vbTab & DeclarationProperty(code_property) & vbCrLf

        ' Look for the property get and set procedures.
        For i As Integer = 1 To code_property.Children.Count
            Dim code_function As CodeFunction2 = DirectCast(code_property.Children.Item(i), CodeFunction2)
            If code_function.FunctionKind = vsCMFunction.vsCMFunctionPropertySet Then
                ' This is the property set procedure.
                Dim params As String = code_function.Prototype( _
                    vsCMPrototype.vsCMPrototypeParamNames Or _
                    vsCMPrototype.vsCMPrototypeParamTypes Or _
                    vsCMPrototype.vsCMPrototypeType)
                params = params.Substring(params.IndexOf("("))
                txt &= vbTab & vbTab & "Set" & params & vbCrLf
                txt &= vbTab & vbTab & vbTab & object_name & "." & code_function.Name & " = value" & vbCrLf
                txt &= vbTab & vbTab & "End Set" & vbCrLf
            ElseIf code_function.FunctionKind = vsCMFunction.vsCMFunctionPropertyGet Then
                ' This is the property get procedure.
                txt &= vbTab & vbTab & "Get" & vbCrLf
                txt &= vbTab & vbTab & vbTab & "Return " & object_name & "." & code_function.Name & vbCrLf
                txt &= vbTab & vbTab & "End Get" & vbCrLf
            Else
                Stop
            End If
        Next i

        txt &= vbTab & "End Property" & vbCrLf & vbCrLf
        Return txt
    End Function

    Private Function ImplementationForVariable(ByVal object_name As String, ByVal code_element As CodeElement2) As String
        Dim txt As String = ""

        ' Start the property procedure.
        Dim code_variable As CodeVariable2 = DirectCast(code_element, CodeVariable2)
        txt &= vbTab & AccessName(code_variable.Access) & " Property "
        txt &= code_variable.Name & "() As " & code_variable.Type.AsString & vbCrLf

        ' Property Get.
        txt &= vbTab & vbTab & "Get" & vbCrLf
        txt &= vbTab & vbTab & vbTab & "Return " & object_name & "." & code_variable.Name & vbCrLf
        txt &= vbTab & vbTab & "End Get" & vbCrLf

        ' Property Set.
        txt &= vbTab & vbTab & "Set(ByVal value As " & code_variable.Type.AsString & ")" & vbCrLf
        txt &= vbTab & vbTab & vbTab & object_name & "." & code_variable.Name & " = value" & vbCrLf
        txt &= vbTab & vbTab & "End Set" & vbCrLf

        txt &= vbTab & "End Property" & vbCrLf & vbCrLf
        Return txt
    End Function
#End Region ' Implementations for code elements.
End Class